
#include <flann/flann.hpp>
#include <string>

#include <stdio.h>

#include "utility.hpp"

typedef float DataType;

using namespace flann;
using namespace std;
using namespace utility;

int main(int argc, char** argv) {
  string ssvfile = "";
  string indexfile = "";
  string indextype = "auto";
  string autoparam = "0.8:0.00:0.01";
  bool help = false;

  int result;
  while ((result = getopt(argc, argv, "i:a:h")) != -1) {
    switch (result) {
      case 'i':
        indextype = optarg;
        break;
      case 'a':
        autoparam = optarg;
        break;
      case 'h':
        help = true;
        break;
    }
  }

  if (argc - optind < 1 || help) {
    cerr << "Usage: ./flann_create [options] train.ssv index" << endl;
    cerr << " -i STRING            Set a index type" << endl;
    cerr << " -a FLOAT:FLOAT:FLOAT Set "
            "TargetPrecision:BuildWeight:SampleFraction" << endl;
    cerr << " -h                   Show this message" << endl;
    cerr << endl;
    cerr << "Example: ./flann_create -i auto -a 0.8:0.0:0.01 train.ssv index"
         << endl;
    exit(-1);
  }
  ssvfile = argv[optind++];
  indexfile = argv[optind++];

  cerr << "data file=" << ssvfile << endl;
  cerr << "index file=" << indexfile << endl;

  size_t dim = 0;
  size_t size = get_data_size(ssvfile, dim);

  cerr << "load data, size=" << size << " dim=" << dim << endl;

  vector<DataType> data;
  data.reserve(size * dim);
  load_ssv(ssvfile, data);
  cerr << dim << ":" << data.size() << endl;

  // matrix(data, rows, columns);
  Matrix<DataType> dataset(&data[0], size, dim);

  Timer timer;
  timer.start();
  cerr << "build begin. " << indextype << endl;
  if (indextype == "linear") {
    Index<L2<DataType> > index(dataset, flann::LinearIndexParams());
    index.buildIndex();
    index.save(indexfile);
  } else if (indextype == "kdtree4") {
    Index<L2<DataType> > index(dataset, flann::KDTreeIndexParams(4));
    index.buildIndex();
    index.save(indexfile);
  } else if (indextype == "kdtree") {
    Index<L2<DataType> > index(dataset, flann::KDTreeSingleIndexParams());
    index.buildIndex();
    index.save(indexfile);
  } else if (indextype == "hierarchy") {
    Index<L2<DataType> > index(dataset,
                               flann::HierarchicalClusteringIndexParams());
    index.buildIndex();
    index.save(indexfile);
  } else if (indextype == "kmeans") {
    Index<L2<DataType> > index(dataset, flann::KMeansIndexParams());
    index.buildIndex();
    index.save(indexfile);
  } else if (indextype == "auto") {
    vector<string> ptokens;
    split(autoparam, ':', ptokens);
    if (ptokens.size() != 3) {
      cerr << "Error: not enough parameter" << endl;
      exit(1);
    }
    float targetPrecision = atof(ptokens[0].c_str());
    float buildWeight = atof(ptokens[1].c_str());
    float sampleFraction = atof(ptokens[2].c_str());
    cerr << "tp:bw:sf=" << targetPrecision << ":" << buildWeight << ":"
         << sampleFraction << endl;
    cerr << "default=0.9, 0.01, 0.1" << endl;
    flann::log_verbosity(flann::FLANN_LOG_DEBUG);
    Index<L2<DataType> > index(
        dataset, flann::AutotunedIndexParams(targetPrecision, buildWeight, 0.0,
                                             sampleFraction));
    index.buildIndex();
    index.save(indexfile);
  } else {
    cerr << "indextype is invalid! " << indextype << endl;
  }
  timer.stop();

  cerr << "build end " << timer << endl;

  return 0;
}
